% A feladvny:

% Egy szakadkon egy hossz s keskeny pall vel t, amely egyszerre
% legfeljebb kt embert br el. A pall egyik oldaln ll ngy ember: 10, 20,
% 50 s 100 vesek. Az N ves ember N perc alatt tud tmenni a hdon. Ha
% ketten mennek t akkor a lassabb embernek megfelel id alatt rnek t.
% 
% Stt van, vilgts nlkl lehetetlen trni, de a trsasgnak csak
% egyetlen zseblmpja van. A feladat: megszervezni az tkelst gy, hogy a
% teljes trsasg a lehet legrvidebb id alatt tkeljen a
% tloldalra. 
% 
% Krds: mennyi id alatt tudnak leggyorsabban tkelni?


:- use_module(library(lists), [member/2, max_list/2]).

% tmegy(Emberek, Lpsek, Id): Az Emberek lista tartalmazza az tkelni
% kvn emberek kort (azaz a palln val tmensi idejt).  Id a
% feladvny legrvidebb megoldsi ideje, Lpsek pedig a legrvidebb idej
% tmenethez szksges lpsek listja.
% :- pred tmegy(list(integer)::in, list(lps)::out, integer::out).
tmegy(Emberek, Lpsek, Id) :-
	length(Emberek, N),
	max_list(Emberek, Max),
	MaxId is (1<<(N+1))*Max,  % 2**(N+1) llapot lehetsges 
	vghelyzet(Emberek, b, Kezdllapot),
	vghelyzet(Emberek, j, Vgsllapot),
	tvonal(Kezdllapot, Vgsllapot, MaxId, Lpsek, Id).

% Egy llapot tartalmazza annak megadst, hogy a lmpa melyik
% oldalon van (b/j), s az egyes emberek melyik oldalon vannak.

% :- type llapot   ---> lmpahely-list(emberhely).
				% Hol a lmpa s hol vannak az emberek?
% :- type lmpahely ==   oldal.
				% A lmpa az egyik oldalon lehet.
% :- type emberhely ---> ember-oldal.
				% Minden adott kor ember mellett
				% ott a tartzkodsi helye.
% :- type ember     ==   integer.
				% Az embereket a koruk jelli.
% :- type oldal     ---> b ; j. % ball illetve jobb oldal.


% :- type lps ---> tmenk>oldal.
				% az adott oldalRA tmennek az tmenk.
% :- type tmenk == ember \/ {ember+ember}.
				% egy vagy kt adott kor ember.

% vghelyzet(Emberek, Oldal, llapot): llapot az a vgllapot, amely
% megfelel az Emberek listnak s az Oldal oldalnak.
% :- pred vghelyzet(list(ember)::in, oldal::in, llapot::out).
vghelyzet(Emberek, Oldal, Oldal-Emberhelyek) :-
	prost(Emberek, Oldal, Emberhelyek).

% prost(L0, Y, L): Az L lista elemei X-Y prok,
% ahol X befutja az L listt.
% :- pred prost(list(ember), oldal, list(emberhely)).
prost([], _, []).
prost([X|XL], Y, [X-Y|XYL]) :-
	prost(XL, Y, XYL).

% tvonal(Kezd, Vgs, Max, t, Id): t egy a Kezd llapotbl a Vgs 
% llapotba vezet Id sszidej lps-sorozat, ahol Id minimlis
% s Id =< Max.
% :- pred tvonal(llapot::in, llapot::in, integer::in,
%                   list(lps)::out, integer::out).
tvonal(Kezd, Vgs, Max, t, Id) :-
	tvonal_6(Kezd, Vgs, Max, _, Nyeresg),
	Nyeresg > 0, !,
	jMax is Max - Nyeresg,
	tvonal(Kezd, Vgs, jMax, t, Id).
tvonal(Kezd, Vgs, Max, t, Max) :-
	tvonal_6(Kezd, Vgs, Max, t, 0).

% tvonal_6(A, B, Max, Lk, Nyer): A s B kztt van egy Max-nl 
% Nyer-rel rvidebb krmentes t (Nyer >= 0), amelynek lpssorozata Lk.
% :- pred tvonal_6(llapot::in, llapot::in, integer::in,
%                   list(lps)::out, integer::out).
tvonal_6(Honnan, Hov, Max, t, Nyer)  :-
	tvonal_6(Honnan, Hov, Max, [], t, Nyer).

% tvonal_6(A, B, Max, Kk, Lk, Nyer): A s B kztt van egy Max-nl 
% Nyer-rel rvidebb krmentes t (Nyer >= 0), amelynek lpssorozata Lk,
% s amely nem megy t a Kk listban szerepl kizrt llapotok egyikn sem.
% :- pred tvonal_6(llapot::in, llapot::in, integer::in, list(llapot)::in,
%                   list(lps)::out, integer::out).
tvonal_6(Hov, Hov, Max, _, [], Max)  :- 
	Max >= 0.
tvonal_6(Honnan, Hov, Max0, Kizk, [Lp|LpL], Nyer) :-
        Max0 > 0,
	tszakasz(Honnan, Kzben, H1, Lp),
	\+ member(Kzben, Kizk),
	Max1 is Max0-H1, 
        tvonal_6(Kzben, Hov, Max1, [Kzben|Kizk], LpL, Nyer).

% tszakasz(ll0, ll1, Id, Lps): A Lps lpssel Id alatt az ll0
% llapotbl az ll1 llapotba lehet jutni.
% :- pred tszakasz(llapot::in, llapot::out, integer::out, lps::out).
tszakasz(Innen-Helyzet0, Ide-Helyzet, Id, E1>Ide) :-     % E1 megy t.
	msik(Innen, Ide),     % A lmpa az Innen oldalon van, Ide a tloldal.
	cserl(E1-Innen, Helyzet0, E1-Ide, Helyzet),       % E1 tmegy
	Id = E1.
tszakasz(Innen-Helyzet0, Ide-Helyzet, Id, E1+E2>Ide) :-  % E1+E2 megy t.
	msik(Innen, Ide),
	cserl(E1-Innen, Helyzet0, E1-Ide, Helyzet1),      % E1 tmegy     
	cserl(E2-Innen, Helyzet1, E2-Ide, Helyzet),       % E2 tmegy
	E1 < E2, Id = E2.


% msik(Egyik, Msik): Egyik s Msik kt klnbz oldal.
% :- pred msik(oldal, oldal).
msik(b, j).
msik(j, b).

% cserl(X, XL, Y, YL): egy X elemet XL-ben Y-ra cserlve kapjuk YL-t.
% Deklaratvabban: X ugyanannyiadik eleme XL-nek, mint Y YL-nek, s a kt
% lista csak ebben az elemben klnbzik. 
% :- pred cserl(T, list(T)::in, T, list(T)::out).
% :- pred cserl(T, list(T)::out, T, list(T)::in).
cserl(X, [X|Xlista], Y, [Y|Xlista]).
cserl(X, [Head|Xlista], Y, [Head|Ylista]) :-
	cserl(X, Xlista, Y, Ylista).

end_of_file.

Futsi pldk:

| ?- tmegy([10,20,50,100], Lpsek, Id).
Id = 170,
Lpsek = [10+20>j,10>b,50+100>j,20>b,10+20>j] ? ;
Id = 170,
Lpsek = [10+20>j,20>b,50+100>j,10>b,10+20>j] ? ;
no
| ?- tmegy([10,20,50,70,100], Lpsek, Id).
Id = 230,
Lpsek = [10+20>j,10>b,10+50>j,10>b,70+100>j,20>b,10+20>j] ? ;
Id = 230,
Lpsek = [10+20>j,10>b,10+50>j,20>b,70+100>j,10>b,10+20>j] ? ;
Id = 230,
Lpsek = [10+20>j,10>b,70+100>j,20>b,10+20>j,10>b,10+50>j] ? ;
Id = 230,
Lpsek = [10+20>j,10>b,70+100>j,20>b,10+50>j,10>b,10+20>j] ? ;
Id = 230,
Lpsek = [10+20>j,20>b,70+100>j,10>b,10+20>j,10>b,10+50>j] ? ;
Id = 230,
Lpsek = [10+20>j,20>b,70+100>j,10>b,10+50>j,10>b,10+20>j] ? ;
Id = 230,
Lpsek = [10+50>j,10>b,10+20>j,10>b,70+100>j,20>b,10+20>j] ? ;
Id = 230,
Lpsek = [10+50>j,10>b,10+20>j,20>b,70+100>j,10>b,10+20>j] ? ;
no
| ?-
